/*****************************************************************************
    readconfig.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2005 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#ifndef READCONFIG_C
#define READCONFIG_C

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <arpa/tftp.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <math.h>

#define CPROPLIST
#include "params.h"
#define	 GLOBDEF
#include "globals.h"
#include "patstr.h"
#undef   GLOBDEF
#ifdef FILE_CACHE
#include "cache.h"
#endif /* FILE_CACHE */
#include "sigmask.h"
#include "util.h"
#include "dll.h"
#include "dllapps.h"
#include "access.h"
#include "logfun.h"
#include "process.h"
#include "getarg.h"
#include "dump.h"
#include "readconfig.h"


enum command_codes {	CMD_UNKNOWN		,
			CMD_DEFAULT_DIR		,
			CMD_MAX_SYSLOG		,
			CMD_MESSAGE_LEVEL	,
			CMD_ROOT_DIR		,
			CMD_ACCESS_LIST		,
			CMD_RESTRICT		,
			CMD_USER		,
			CMD_GROUP		,
			CMD_LIMIT		,	
			CMD_FILTER		,
			CMD_MAP			,
			CMD_LOGFACILITY		,
			CMD_CREATEMODE		,
			CMD_TIMEOUT		,
			CMD_LOCALE		,
#ifdef FILE_CACHE	
			CMD_FILECACHE		,
			CMD_CACHEOBJECT		,
#endif /* FILE_CACHE */
			CMD_DEFTIMEOUT		};

typedef struct CMDS_s { char *Name;
                        int Code;
                        int minArgs;
			int intCheck; } CMDS_t;

static CMDS_t configCmds[] = { 
		{ "defaultdirectory", 	CMD_DEFAULT_DIR,	2,	0 } ,
		{ "maxsyslog",	 	CMD_MAX_SYSLOG, 	2,	1 } ,
		{ "messagelevel", 	CMD_MESSAGE_LEVEL, 	2,	1 } ,
		{ "rootdirectory", 	CMD_ROOT_DIR, 		2,	0 } ,
		{ "accesslist", 	CMD_ACCESS_LIST, 	4,	1 } ,
		{ "restrict", 		CMD_RESTRICT, 		3,	1 } ,
		{ "defaultTiming",	CMD_DEFTIMEOUT, 	2,	0 } ,
		{ "user", 		CMD_USER, 		2,	0 } ,
		{ "group", 		CMD_GROUP, 		2,	0 } ,
		{ "limit", 		CMD_LIMIT, 		2,	1 } ,
		{ "filter", 		CMD_FILTER, 		3,	1 } ,
		{ "map", 		CMD_MAP, 		4,	1 } ,
		{ "logfacility",	CMD_LOGFACILITY,	2,	0 } ,
		{ "createmode",		CMD_CREATEMODE,		2,	1 } ,
		{ "locale",		CMD_LOCALE,		2,	0 } ,
		{ "timerlist",		CMD_TIMEOUT,		3,	1 } ,
#ifdef FILE_CACHE	
		{ "filecache",		CMD_FILECACHE,		2,	0 } ,
		{ "cacheobject",	CMD_CACHEOBJECT,	2,	0 } ,
#endif /* FILE_CACHE */
		{ NULL, 		CMD_UNKNOWN, 		0,	0 } };

static int curLine;

/*
	config error prefix
*/

static char *configErr(void)
{
    static char confErrString[64];

    if (curLine)
        sprintf(confErrString, _("Config line %d:"), curLine);
    else
        strcpy(confErrString, _("Config error:"));

    return confErrString;
}

/*
	split config line into chunks a'la argc / argv
*/

static int cutConfigLine (char *line, int *argc, char **argv, 
                          int maxargs, char **ctx)
{
    static char *sep = " \t", *quo = "''\"\"", *tok;
    int i=0;

    for (tok = qstrtok(line, sep, ctx, quo); tok != NULL; 
         tok = qstrtok(NULL, sep, ctx, quo)) {
        if (i >= maxargs) {
            LOG_MSG1 (LOG_ERR, _("%s too many arguments (> %d)"), 
                      configErr(), maxargs);
            return -1;
        }
        argv[i++] = tok;
    }
    *argc = i;
    return i;
}

/*
	extended string compare
	- case insensitive
	- ignores embedded dashes and underscores
*/

static int cmpXtended (char *s1, char *s2)
{
    char c1, c2;

    for (; c1 = *s1, c2 = *s2; s1++, s2++) {
        while (c1 == '-' || c1 == '_')
    	    c1 = *(++s1);

        if (isupper (c1))
	    c1 = tolower (c1);

        while (c2 == '-' || c2 == '_')
	    c2 = *(++s2);

        if (isupper (c2))
	    c2 = tolower (c2);

        if (c1 != c2 || c1 == '\0')
	    break;
    }
    return (int) (c1 - c2);
}

/*
	evaluate timing parameters of an already broken off command string
*/

static int parseNTimes(char **argv, int argc, tout_t *tp)
{
    int i, rtc, sts=0;

    for (i=0; i<argc; i++) {
        if ((rtc = parseTimes(argv[i], tp)) < 0)
            return -1;
        sts += rtc;
    }
    return sts;
}

/*
	read a line from the config file
	deal with continuation constructs and
	delete comments
*/

static char *getConfigLine(FILE *fd)
{
#define intSIZ 8192
    static char line[intSIZ], *ptr, *start;
    int rest, len;

    for (ptr = line, rest = intSIZ; fgets(ptr, rest, fd) != NULL;) {
        curLine++;
        for (start = ptr; *start == ' ' || *start == '\t'; start++);
        if (*start == '#') continue;
        len = strlen(start) - 1;
        if (len < 1) continue;
        if (start > ptr)
            memmove(ptr, start, len);
        ptr += len;
        if (*(ptr-1) != '\\') break;
        len--;
        ptr--;
        rest -= len;
    }
    *ptr = '\0';
    return ptr == line ? NULL : line;
}

/*
	scan the command list for the current token
*/

static CMDS_t *findCmd (char *token)
{
    CMDS_t *cp;

    for (cp = configCmds; cp->Name != NULL; cp++)
        if (!cmpXtended (cp->Name, token)) 
  	    break;

    return cp;
}


/*
	open the dump file
*/

FILE *openDump()
{
    FILE *dumpFILE;
    char *txdumpFile;

    if (spMode || checkOnly)
        dumpFILE = stderr;
    else {
        txdumpFile = xlatePath(dumpFile);
        dumpFILE = fopen (txdumpFile,"w+");
        if (dumpFILE == NULL)
            LOG_MSG1 (LOG_ERR, _("Cannot open dumpfile '%s'"), txdumpFile);
        Free (txdumpFile);
    }
    return dumpFILE;
}
           
/*
	close the dump file
*/

static void closeDump(FILE *dumpFILE)
{
    if (dumpFILE != stderr && dumpFILE != NULL)
        fclose (dumpFILE);
}

/*
	dump the parameters derived from the command line
*/
static void cmdDump (FILE *dump)
{
    static char *Yes, *No;

    Yes = _("Yes");
    No  = _("No");

    if (dump == NULL)
        return;

    fprintf (dump,_("\nCommand line options for '%s':\n\n"
            "Message Level      (-m): [%c] %d\n"
            "Max SYSLOG Level   (-x): [%c] %d\n"
            "Enable Broadacst   (-b): [%c] %s\n"
            "Single Process     (-s): [%c] %s\n"
            "Check only         (-c): [%c] %s\n"
            "Safe Log           (-S): [%c] %s\n"
            "Silent Deny        (-q): [%c] %s\n"
            "Transfer Options   (-o): [%c] %s\n"
            "Core dump enabled  (-e): [%c] %s\n"
            "ACK wait on err    (-E): [%c] %s\n"
            "XFER-Recovery      (-R): [%c] %s\n"
            "Show Help          (-?): [%c] %s\n"
            "resolve host names (-n): [%c] %d\n"
            "Log Facility       (-f): [%c] %s\n"
            "Local Translation  (-A): [%c] %s\n"
            "Default Timing     (-t): [%c] %s\n"
            "Local address      (-l): [%c] %s\n"
            "Local port         (-p): [%c] %d\n"
            "Config File        (-C): [%c] %s\n"
            "Dump File          (-D): [%c] %s\n"
            "Log File           (-L): [%c] %s\n"
            "PID File           (-P): [%c] %s\n"),
            myName,
            shortMF(MF_msgLevel),    msgLevel,
            shortMF(MF_maxSyslog),   maxSyslog, 
            shortMF(MF_broadCast),   broadCast     ? Yes : No,
            shortMF(MF_spMode),      spMode        ? Yes : No,
            shortMF(MF_checkOnly),   checkOnly     ? Yes : No,
            shortMF(MF_safeLog),     safeLog       ? Yes : No,
            shortMF(MF_silentDeny),  silentDeny    ? Yes : No,
            shortMF(MF_xferOptions), xferOptions   ? Yes : No,
            shortMF(MF_coreDump),    coreDump      ? Yes : No,
            shortMF(MF_errAck),      errAck        ? Yes : No,
            shortMF(MF_allowRecover),allowRecover  ? Yes : No,
            shortMF(MF_Help),        Help          ? Yes : No,
            shortMF(MF_doResolv),    doResolv,
            shortMF(MF_logFac),      getFacilityByValue(logFac),
            shortMF(MF_doLocale),    (doLocale==NULL) ? "-" : 
                                         ((*doLocale=='\0') ? 
                                             getenv("LC_ALL") : doLocale),
            shortMF(MF_defTimer),    timer2asc(&defTimer, T2A_ALL),
            shortMF(MF_serverAddr),  xinet_ntoa(serverAddr, doResolv),
            shortMF(MF_serverPort),  serverPort, 
            shortMF(MF_configFile),  configFile,
            shortMF(MF_dumpFile),    dumpFile == NULL ? "-" : dumpFile,
            shortMF(MF_logFile),     logFile  == NULL ? "-" : logFile,
            shortMF(MF_pidFile),     pidFile);
}

/*
	dump the current global configuration parameters
*/

static void configDump(FILE *dump)
{
    struct passwd *pp;
    struct group  *gp;
    static char *unk = "<unknown>";

    if (dump == NULL)
        return;

    pp = getpwuid(childUid);
    gp = getgrgid(childGid);

    fprintf(dump, _("\n"
                "Message Level    : [%c] %d\n"
                "Max SYSLOG Level : [%c] %d\n"
                "Default Timing   : [%c] %s\n"
                "User             : [%c] %-5d %s\n"
                "Group            : [%c] %-5d %s\n"
                "Default Dir.     : [%c] %s\n"
                "Root Directory   : [%c] %s\n"
                "Create Mode      : [%c] %o\n"),
                shortMF(MF_msgLevel),  msgLevel, 
                shortMF(MF_maxSyslog), maxSyslog, 
                shortMF(MF_defTimer),  timer2asc(&defTimer, T2A_ALL), 
                shortMF(MF_childUid),  childUid, pp==NULL ? unk : pp->pw_name, 
                shortMF(MF_childGid),  childGid, gp==NULL ? unk : gp->gr_name, 
                shortMF(MF_defaultDir),defaultDir == NULL ? "-" : defaultDir, 
                shortMF(MF_rootDir),   rootDir    == NULL ? "-" : rootDir, 
                shortMF(MF_rootDir),   createMode);
}

/*
	dump info on open file descriptors
*/

void fdDump(FILE *dump)
{
    int fd, cnt, soc, coe;
    struct stat statBuf;

    if (dump == NULL)
        return;

    for (fd = 0, cnt = 0; fd < 32; fd++) {
        if (fstat(fd, &statBuf)) 
            continue;
        if (!cnt++) {
            fprintf (dump, _("\nOpen file descriptors:\n"
                               "Descr.\tSocket\tClose-on-exec\n"));
        }
        soc = isaSocketFd (fd);
        coe = fcntl (fd, F_GETFD);
        if (coe < 0)
            coe = 0;
        fprintf (dump,"%d\t%s\t%s\n",
                 fd, soc ? _("yes") : _("no"), coe ? _("yes") : _("no"));
    }
}

/*
	dump configuration and access lists
*/

void genDump(int what)
{
    time_t itime;
    FILE *dump;

    time(&itime);
    if((dump = openDump()) == NULL)
        return;
    version(dump, "server");
    copyright(dump, 1);
    fprintf(dump, _("Configuration dump generated by PID %d at %s"),
            getpid(), asctime(localtime(&itime)));
    cmdDump(dump);
    configDump(dump);
#ifdef FILE_CACHE
    cacheDump(dump, what);
#endif /* FILE_CACHE */
    fdDump(dump);
    accessDump(dump);
    timerDump(dump);
    filterDump(dump);
    limitsDump(dump);
    closeDump(dump);
}

int resetDir (char **var, char *dir, int flag)
{
    if (*var != NULL) {
        if ((dir == NULL) || strcmp(*var, dir) || (flag)) {
            Free(*var);
            *var = NULL;
        }
    }
            
    if ((*var == NULL) && !flag && !dirAccess(dir, 0))
        if ((*var = Strdup(dir)) == NULL)
            return 1;
    return 0;
}

static int resetConfig ()
{
    globReset(MF_CONFIG);

    if (restrictListH != NULL)
        deleteReChain (restrictListH);
    if ((restrictListH = initReChain()) == NULL)
        return -1;

    if (accessListH != NULL)
        deleteACLChain (accessListH);
    if ((accessListH = initACLChain()) == NULL)
        return -1;

    if (filterListH != NULL)
        deleteFilterChain (filterListH);
    if ((filterListH = initFilterChain()) == NULL)
        return -1;

    if (mapListH != NULL)
        deleteReChain (mapListH);
    if ((mapListH = initReChain()) == NULL)
        return -1;

    if (threadListH != NULL)
        deleteACLChain (threadListH);
    if ((threadListH = initACLChain()) == NULL)
        return -1;
 
    if (timerListH != NULL)
        deleteTimerChain (timerListH);
    if ((timerListH = initTimerChain()) == NULL)
        return -1;

/*
    if (procListH != NULL) 
        SIGBLOCK(SIGCHLD, (deletePidChain (procListH)));
*/
    if (procListH == NULL) 
        if ((procListH = initPidChain()) == NULL)
            return -1;

#ifdef FILE_CACHE

    if (cacheListH != NULL)
        deleteCacheChain (cacheListH);
    if ((cacheListH = initCacheChain()) == NULL)
        return -1;

    if (objListH != NULL)
        deleteReChain (objListH);
    if ((objListH = initReChain()) == NULL)
        return -1;

#endif /* FILE_CAHE */

    changeRoot = ((rootDir != NULL) && !spMode);

    return 0;
}


/* 
	(re)-read the tftpd config file.  
	We keep an open handle on the file
	and check to see if it's been modified since the last read.
*/
static int readconfigFile2(FILE *configFP)
{
#define MAXARGS 64

    char *line;
    CMDS_t *cmdp;
    int perm, list;
    int cargc;
    char *cargv[MAXARGS];
    char *ctx;
    int rtc = 0;
    int fnum;
    int fac;
    int dtype, coffset;
    tout_t times;
    char *mask;
    u_long maxsize;
    int i;
#ifdef FILE_CACHE
    ulong maxC, maxO;
    int flags;
#endif /* FILE_CACHE */

/* 
	Read the config file 
*/
    LOG_MSG4 (LOG_INFO, _("Processing config file '%s'"),configFile);

    curLine = 0;

    if (resetConfig ()) {
        errNoDynmem;
        return 1;
    }

    while ((line = getConfigLine(configFP)) != NULL) {

	if (cutConfigLine (line, &cargc, cargv, MAXARGS, &ctx) < 0)
            continue;

        cmdp = findCmd (cargv[0]);

/*
	check config clause for sufficient number of parameters
*/
        if (cargc < cmdp->minArgs) {
            LOG_MSG1 (LOG_ERR,
                     _("%s not enough arguments to '%s' directive (need %d)"),
                     configErr(), cargv[0], cmdp->minArgs);
            rtc++;
            if (checkOnly)
                continue;
            else
                return 1;
        }

/*
	check numeric argument if necessary
*/
        if (cmdp->intCheck) {
            char *fmt;
            if (!strncmp(cargv[1], "0x", 2))
                fmt = "0x%x";
            else if (!strncmp(cargv[1], "0X", 2))
                fmt = "0X%x";
            else if (!strncmp(cargv[1], "0", 1))
                fmt = "%o";
            else
                fmt = "%u";
            if (sscanf(cargv[1], fmt, (unsigned *)&list) != 1) {
                LOG_MSG1 (LOG_ERR, 
                        _("%s illegal numeric argument '%s' in '%s' directive"),
                        configErr(), cargv[1], cargv[0]);
                rtc++;
                if (checkOnly)
                    continue;
                else
                    return 1;
            }
        }

        switch (cmdp->Code) {

	    case CMD_DEFAULT_DIR:	/* default directory */

                if (*cargv[1] == '!') {
                    keepBelow = 1;
                    cargv[1]++;
                }

                if (*cargv[1] != '/') {
                    LOG_MSG2 (LOG_WARNING,
                              _("No absolute path in DefaultDirectory '%s'"),
                               cargv[1]);
                }
                if (!globCheckCfg(defaultDir)) {
                    if (dirAccess(cargv[1], R_OK|W_OK|X_OK)) {
                        LOG_MSG1 (LOG_ERR, _("Default directory '%s': %m"),
                                  cargv[1]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                    }
                    strcpy(defaultDir, cargv[1]);
                }
                break;


            case CMD_ROOT_DIR:		/* virtual root directory */

                if (spMode) {
                    LOG_MSG1 (LOG_ERR, 
                              _("'rootDirectory' invalid in Single-process mode -- ignored"));
                    break;
                }

                if (*cargv[1] != '/') {
                    LOG_MSG1 (LOG_ERR, 
                              _("No absolute path in RootDirectory '%s'"), 
                               cargv[1]);
                    rtc++;
                    if (!checkOnly)
                        return 1;
                    break;
                }

                if (!globCheckCfg(rootDir)) { 
                    if (dirAccess(cargv[1], R_OK|W_OK|X_OK)) {
                        LOG_MSG1 (LOG_ERR, _("Root directory '%s': %m"),
                                  cargv[1]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                    }
                    if ((rootDir = Strdup(cargv[1])) == NULL) {
                        errNoDynmem;
                        return 1;
                    }
                    changeRoot = 1;
                }
                break;


            case CMD_MAX_SYSLOG:	/* set max. syslog message level */

                globCfg(maxSyslog,list);
                break;


            case CMD_MESSAGE_LEVEL:	/* set message level */

                globCfg(msgLevel,list);
                break;


            case CMD_LOGFACILITY:	/* set log facility */

                fac = getFacilityByName(cargv[1]);
                
                if (fac < 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal Log Facility '%s'"),
                              configErr(), cargv[1]);
                    rtc++;
                    if (!checkOnly)
                        return 1;
                }
                globCfg(logFac, fac);
                break;

            case CMD_CREATEMODE:	/* set createmode */

                globCfg(createMode,(list & 0777));
                break;


            case CMD_TIMEOUT:

                times = defTimer;
                if (parseNTimes(&cargv[2], cargc-2, &times) <= 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal timeout list %d"), 
                              configErr(), list);
                    rtc++;
                    if (!checkOnly)
                        return 1;
                }
                else { 
                    (void)initBackoff(&times, STRENGTH);

                    switch(addTimer(timerListH, list, &times, NULL)) {

                        case -1:
                            errNoDynmem;
                            return 1;

                        case -2:
                            LOG_MSG1 (LOG_ERR, _("%s duplicate timerlist %d"),
                                      configErr(), list);
                            rtc++;
                            if (!checkOnly)
                                return 1;
                            break;
                    }
                }
                break;

            case CMD_DEFTIMEOUT:

                times = defTimer;
                if (parseNTimes(&cargv[2], cargc-2, &times) <= 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal default timing"), 
                              configErr());
                    rtc++;
                    if (!checkOnly)
                        return 1;
                }
                else {
                    globCfg(defTimer, times);
                }
                break;

            case CMD_ACCESS_LIST:	/* create or extend an access list */

                if ((perm = evalPermission(cargv[2])) < 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal permission list '%s'"),
                              configErr(), cargv[2]);
                    rtc++;
                    if (!checkOnly)
                        return 1;
                }

                if (cargc < 5) {	/* no netmask field */
                    if ((mask = cidrsplit(cargv[3])) == NULL) {
                        LOG_MSG1 (LOG_ERR, _("%s illegal network address: %d"),
                                  configErr(), cargv[3]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                    }
                }
                else {
                    mask = cargv[4];
                }

                if (list <= 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal ACL number: %d"),
                              configErr(), list);
                    rtc++;
                    if (!checkOnly)
                        return 1;
                    break;
                }

                switch (addACL(accessListH,list,perm,cargv[3],mask,
                               doResolv & RES_CONFIG, NULL)) {

                    case -1:
                        errNoDynmem;
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, _("%s unresolvable hostname '%s'"),
                                  configErr(), cargv[3]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                }
                break;


            case CMD_RESTRICT:	/* restrict file access by access list */


                if (cargc > 4) {
                    if (sscanf(cargv[4], "%u", (unsigned *)&fnum) != 1) {
                        LOG_MSG1 (LOG_ERR, 
                                  _("%s illegal filter number '%s' in 'restrict' directive"),
                                           configErr(), cargv[4]);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                    }
                }
                else {
                    fnum = 0;
                }

                if (cargc > 3) {
                    if (sscanf(cargv[3], "%lu",&maxsize) != 1) {
                        LOG_MSG1 (LOG_ERR, 
                                  _("%s illegal transfer size limit '%s' in 'restrict' directive"),
                                           configErr(), cargv[3]);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                    }
                }
                else {
                    maxsize = 0;
                }

                switch (addRePattern(restrictListH, list, cargv[2], NULL,
                                     fnum, RT_RRQ|RT_WRQ, maxsize, NULL)) {

                    case -1:
                        errNoDynmem;
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, _("%s illegal regular expression '%s'"),
                                  configErr(), cargv[2]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                }

                if (getACLByMark(accessListH, NULL, list) == NULL)
                    LOG_MSG2 (LOG_WARNING,
                              _("%s reference to undefined ACL '%d'"),
                              configErr(), list);

                if (fnum && (getFilterByMark(filterListH, NULL, fnum) == NULL))
                    LOG_MSG2 (LOG_WARNING,
                              _("%s reference to undefined filter '%d'"),
                              configErr(), fnum);
                break;


            case CMD_USER:		/* set new user */

                {
                    uid_t uid;
                    if (txt2uid(cargv[1], &uid)) {
                        rtc++;
                        if (!checkOnly)
                            return 1;
                    }
                    else
                        globCfg(childUid,uid);
                }
                break;


	    case CMD_GROUP:		/* set new group */

                {
                    gid_t gid;
                    if (txt2gid(cargv[1], &gid)) {
                        rtc++;
                        if (!checkOnly)
                            return 1;
                    }
                    else
                        globCfg(childGid,gid);
                }
                break;


            case CMD_LIMIT:	/* add a group limit */

                if (cargc < 4) {
                    if ((mask = cidrsplit(cargv[2])) == NULL) {
                        LOG_MSG1 (LOG_ERR, _("%s illegal network address: %s"),
                                  configErr(), cargv[2]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                    }
                }
                else {
                    mask = cargv[3];
                }

                if (!list) {
                    LOG_MSG1 (LOG_ERR, _("%s limit list no. must be > 0"),
                              configErr());
                    rtc++;
                    if (!checkOnly)
                        return 1;
                    break;
                }

                switch (addACL(threadListH, -1, list, cargv[2],
                               mask, doResolv & RES_CONFIG, NULL)){

                    case -1:
                        errNoDynmem;
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, _("%s unresolvable hostname '%s'"),
                                  configErr(), cargv[3]);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                }
                break;


            case CMD_FILTER:	/* add a filter description */

                if (list <=0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal filter %d"),
                                        configErr(), list);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                }

                if (getFilterByMark(filterListH, NULL, list) != NULL) {
                    LOG_MSG1 (LOG_ERR, _("%s duplicate filter declaration %d"),
                                        configErr(), list);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                }

                for (i = 2; i < cargc; i++) {
                    if (checkFileMap(cargv[i])) {
                        LOG_MSG1 (LOG_ERR, 
                                  _("%s illegal filter declaration '%s'"),
                                  configErr(), cargv[i]);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                    }
                }

                if (addFilter(filterListH,list,cargv[2],cargv[3],NULL)) {
                    errNoDynmem;
                    return 1;
                }
                break;


            case CMD_MAP:	/* file map filter entry	*/

                if (checkFileMap(cargv[3])) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal file map '%s'"),
                                        configErr(), cargv[3]);
                        rtc++;
                        if (checkOnly)
                            continue;
                        else
                            return 1;
                }

                switch (*cargv[3]) {
                    case '>':
                        dtype = RT_WRQ;
                        coffset = 1;
                        break;
                    case '<':
                        dtype = RT_RRQ;
                        coffset = 1;
                        break;
                    default:
                        dtype = RT_RRQ | RT_WRQ;
                        coffset = 0;
                        break;
                }                

                if (addRePattern(mapListH, list, cargv[2], cargv[3]+coffset,
                                 0 ,dtype, 0, NULL)) {
                    errNoDynmem;
                    return 1;
                }
                break;

            case CMD_LOCALE:	/* set translation locale */

                globCfg(doLocale, Strdup(cargv[1]));
                initLocale(doLocale);
                break;

#ifdef FILE_CACHE	
            case CMD_FILECACHE:	/* enable file cache */

                if (cvtunit(cargv[1], &maxC) || 
                    cvtunit(cargv[2], &maxO)) {
                    LOG_MSG1 (LOG_ERR, _("%s invalid cache parameter"),
                              configErr());
                    rtc++;
                    if (checkOnly)
                        continue;
                    else
                        return 1;
                }
                globCfg(maxFilecache, maxC);
                globCfg(maxCacheobj,  maxO);
                break;

            case CMD_CACHEOBJECT: /* enter object(s) into the cache list */

                flags = patstr(cproplist, cargv[2], ",", 
                               PAT_EXACT|PAT_NOCASE|PAT_MULTI);
                if (flags < 0) {
                    LOG_MSG1 (LOG_ERR, _("%s illegal cache object properties"),
                              configErr());
                    rtc++;
                    if (checkOnly)
                        continue;
                    else
                        return 1;
                }
                if (!(flags & CAC_MODE))
                    flags |= CAC_BINARY;

                if ((*cargv[1] == '/') || (defaultDir == NULL)) {
                    if (addRePattern(objListH,0,cargv[1],NULL,flags,0,0,NULL)) {
                        errNoDynmem;
                        return 1;
                    }
                }
                else if ((*cargv[1] == '.') && (*(cargv[1]+1) == '/')) {
                    char *objPath;
                    int objLen;
                    size_t objSize;
                    if ((objPath = getPathBuf(&objSize)) == NULL) {
                        errNoDynmem;
                        return -1;
                    }
                    getcwd(objPath, objSize);
                    objLen = strlen(objPath);
                    if ((size_t)(objLen + strlen(cargv[1])) > objSize) {
                        LOG_MSG1 (LOG_ERR, _("%s path too long"), configErr());
                        Free(objPath);
                        rtc++;
                        if (!checkOnly)
                            return 1;
                        break;
                    }
                    *(objPath+objLen) = '/';
                    strcpy(objPath+objLen+1, cargv[1]+2);
                    if (addRePattern(objListH,0,objPath,NULL,flags,0,0,NULL)) {
                        errNoDynmem;
                        Free(objPath);
                        return 1;
                    }
                    Free(objPath);
                }
                else {
                    char *objPath;
                    if ((objPath=(char *)Malloc(strlen(defaultDir)+
                                                strlen(cargv[1])+2)) == NULL) {
                        errNoDynmem;
                        return -1;
                    }
                    sprintf(objPath, "%s/%s", defaultDir, cargv[1]);
                    if (addRePattern(objListH,0,objPath,NULL,flags,0,0,NULL)) {
                        errNoDynmem;
                        return 1;
                    }
                    Free(objPath);
                }
                break;
#endif /* FILE_CACHE */

            default:		/* illegal configuration options */

                LOG_MSG1 (LOG_ERR, _("%s Illegal configuration keyword '%s'"),
                          configErr(), cargv[0]);
                rtc++;
                if (!checkOnly)
                    return 1;
                break;
	}
    }

    (void)initBackoff(&defTimer, STRENGTH);

    curLine = 0;


/*
	sort acl, filter and timer declarations in ascending order
*/
    sortACLbyIP(accessListH);
    sortFilterByMark(filterListH, DLL_SORTASC);
    sortTimerByMark(timerListH, DLL_SORTASC);

    LOG_MSG4(LOG_INFO, _("Default timing: %s"), 
                        timer2asc(&defTimer,T2A_XFER|T2A_SETUP|T2A_RET));

#ifdef FILE_CACHE

    rtc += addCacheObjs();
    preloadCache();

#endif /* FILE_CACHE */

    if (checkOnly)
          genDump(1);
     
    LOG_MSG4 (LOG_INFO, _("%d configuration errors"), rtc); 
    return rtc;
}

/* 
	Jacket routine for readconfigFile2().

	This one does the pre-checking of configuration changes
	and assures the config file will be correctly closed when
	readconfigFile2() returns from various error exits.
*/
int readconfigFile (int init)
{
    stat_t newStat;
    int changed, rtc;
    FILE *configFP;
/*
	The config file is processed by the parent process only
*/
    if (getpid() != parentPid)
        return 0;
/*
	Check for changes
*/
    if (stat(configFile, &newStat) < 0) {
        LOG_MSG1 (LOG_ERR, _("Cannot get config file status: %m"));
        LOG_MSG5 (LOG_DEBUG, "Working dir: %s, Config file: %s",
                  getcwd(NULL,0), configFile);
        return 1;
    }

    if (!init) {
        changed = (newStat.st_mtime != configStat.st_mtime) ||
                  (newStat.st_ino   != configStat.st_ino  );

        if (changed) {
            LOG_MSG5 (LOG_INFO, _("Config file '%s' changed"), configFile);
            if (!autoConf)
                LOG_MSG5 (LOG_NOTICE, _("Auto-reconfig disabled"));
        }

        if (forceConfig)
            LOG_MSG5 (LOG_NOTICE, _("Reconfig forced"));

        if (!(changed && autoConf) && !forceConfig)
            return 0;
    }

    forceConfig = 0;
    configStat  = newStat;

    if ((configFP = fopen (configFile, "r")) == NULL) {
        LOG_MSG1 (LOG_ERR, _("Cannot open config file '%s': %m"), configFile);
        return 1;
    }

    rtc = readconfigFile2(configFP);

    fclose (configFP);

    return rtc;
}


static void usage()
{
    fprintf(stderr, 
            _("Usage is:\n"
            "%s [-s] [-n] [-m lev] [-c] [-t list] [-S] [-i] [-b][-f fac] [-x]"
            " [-?] [-l addr] [-p port] [-u user] [-g group][-q] [-o] [-e] [-V]"
            " [-N] [-A loc] [-B size] [-C conf] [-D dump] [-L log]\n"
            "\t-s\tRun in single process mode (don't fork)\n"
            "\t-n\tname to address (BIND) lookup (1: config, 2: runtime)\n"
            "\t-m lev\tSet message level to <lev> (default: %d)\n"
            "\t-c\tJust check the configuration and exit\n"
            "\t-t list\tSet default timing parameters\n"
            "\t-S\tAllow secure messages only\n"
            "\t-b\tEnable broadcast receiver mode\n"
            "\t-f fac\tFacility type for syslog messages (default: %s)\n"
            "\t-x max\tLimit syslog messages to priority (default: %d)\n"
            "\t-?\tPrint this text\n"
            "\t-l addr\tlocally bind to this address or hostname\n"
            "\t-p port\tlocally bind to this port number or service\n"
            "\t-u user\trun as <user>\n"
            "\t-g user\trun as <group>\n"
            "\t-q\tSilent access denials (no error packets)\n"
            "\t-o\tenable RFC 1782ff/2347 option processing\n"
            "\t-e\tenable core dump of child processes\n"
            "\t-V\tPrint copyright and version info\n"
            "\t-N\tdo not automatically reconfigure on config file changes\n"
            "\t-A loc\tset translation locale to <loc>\n"
            "\t-C conf\tConfig file (default: '%s')\n"
            "\t-P pid\tPID file (default: '%s')\n"
            "\t-D dump\tOptional Dump file\n"
            "\t-L log\tOptional log file\n"),
            myName, MESSAGELEVEL, LOGFACILITY, MAXSYSLOG, CONFIGFILE, PIDFILE);
}

/*
	process command line options
*/

void evalCommandLine (int argc, char **argv, int merge)
{
    static char *allopts = "a:sn:m:cSx:qboeEu:g:p:l:f:P:C:D:L:?VNA:B:yR";
    char *clogFac="", *cPort="", *cAddr, *cUser=TFTPUSER, *cGroup=TFTPGROUP;
    uid_t uid;
    gid_t gid;

    globReset(MF_CMD|MF_CONFIG);

    getcwd(defaultDir, sizeof(defaultDir));
   
    if (argc) {
        char *p;
        if ((p = strrchr(argv[0], '/')) == NULL)
            myName = argv[0];
        else
            myName = p+1;
    }
    else
        myName = "tftpd";

    if (getarg (argc, argv, allopts, '?', "", NULL) != NULL) {
        version(stderr, "server");
        usage();
        exit(0);
    }

    if (getarg (argc, argv, allopts, 'V', "", NULL) != NULL) {
        version(stderr, "server");
        copyright(stderr, 0);
        showBuild(stderr);
        exit(0);
    }

    cAddr = xinet_ntoa(INADDR_ANY,0);

    if (!globCheckCmd(doLocale))
        doLocale    = getarg(argc,argv,allopts,'A',"",NULL);
    initLocale(doLocale);
    if (!globCheckCmd(accTest))
        accTest     =      getarg(argc,argv,allopts,'a',NULL,NULL);
    if (!globCheckCmd(spMode))
        spMode      = atoi(getarg(argc,argv,allopts,'s',"1", "0"));
    if (!globCheckCmd(noDetach))
        noDetach    = atoi(getarg(argc,argv,allopts,'y',"1", "0"));
    if (!globCheckCmd(checkOnly))
        checkOnly   = atoi(getarg(argc,argv,allopts,'c',"1", "0"));
    if (!globCheckCmd(safeLog))
        safeLog     = atoi(getarg(argc,argv,allopts,'S',"1", "0"));
    if (!globCheckCmd(silentDeny))
        silentDeny  = atoi(getarg(argc,argv,allopts,'q',"1", "0"));
    if (!globCheckCmd(broadCast))
        broadCast   = atoi(getarg(argc,argv,allopts,'b',"1", "0"));
    if (!globCheckCmd(xferOptions))
        xferOptions = atoi(getarg(argc,argv,allopts,'o',"1", "0"));
    if (!globCheckCmd(coreDump))
        coreDump    = atoi(getarg(argc,argv,allopts,'e',"1", "0"));
    if (!globCheckCmd(errAck))
        errAck      = atoi(getarg(argc,argv,allopts,'E',"1", "0"));
    if (!globCheckCmd(allowRecover))
        allowRecover= atoi(getarg(argc,argv,allopts,'R',"1", "0"));
    if (!globCheckCmd(autoConf))
        autoConf    = atoi(getarg(argc,argv,allopts,'N',"0", "1"));
    if (!globCheckCmd(doResolv))
        doResolv    = atoi(getarg(argc,argv,allopts,'n',"3", "1"));
    if (!globCheckCmd(msgLevel))
        msgLevel    = atoi(getarg(argc,argv,allopts,'m',STRING(MESSAGELEVEL),STRING(MESSAGELEVEL)));
    if (!globCheckCmd(maxSyslog))
        maxSyslog   = atoi(getarg(argc,argv,allopts,'x',STRING(MAXSYSLOG),STRING(MAXSYSLOG)));
        cPort       = getarg(argc,argv,allopts,'p',TFTPSERVICE,TFTPSERVICE);
        cAddr       = getarg(argc,argv,allopts,'l',cAddr,cAddr);
        cUser       = getarg(argc,argv,allopts,'u',NULL,NULL);
        cGroup      = getarg(argc,argv,allopts,'g',NULL,NULL);
        clogFac     = getarg(argc,argv,allopts,'f',LOGFACILITY,LOGFACILITY);
    if (!globCheckCmd(pidFile))
        pidFile     = getarg(argc,argv,allopts,'P',PIDFILE,PIDFILE);
    if (!globCheckCmd(configFile))
        configFile  = getarg(argc,argv,allopts,'C',CONFIGFILE,CONFIGFILE);
    if (!globCheckCmd(dumpFile))
        dumpFile    = getarg(argc,argv,allopts,'D',DUMPFILE,DUMPFILE);
    if (!globCheckCmd(logFile))
        logFile     = getarg(argc,argv,allopts,'L',NULL,NULL);

    if (!globCheckCmd(logFac))
        logFac     = getFacilityByName(clogFac);
    if (!globCheckCmd(serverPort))
        serverPort = setPort(cPort);
    if (!globCheckCmd(serverAddr))
        serverAddr = xinet_addr(cAddr, doResolv);
    if (cUser == NULL)
        childUid = geteuid();
    else if ((!txt2uid(cUser, &uid)) && (!globCheckCmd(childUid)))
        childUid = uid;
    if (cGroup == NULL)
        childGid = getegid();
    else if ((!txt2gid(cGroup, &gid)) && (!globCheckCmd(childGid)))
        childGid = gid;

    if (!merge) 
        return;

    if (accTest != NULL) {
        msgLevel  = 4;
        maxSyslog = 0;
        spMode    = 1;
        logPrefix = 0;
	logFile   = NULL;
    }

    if (checkOnly) {
        maxSyslog = 0;
        if(msgLevel < 4)
            msgLevel  = 4;
    }
#   undef CMSGLEVEL
#   undef CMAXSYSLOG
#   undef C_IN_ADDR
}        

#ifdef TEST

int main (int argc, char **argv)
{
    int rtc;
    FILE *dumpFILE = stderr;

    evalCommandLine(argc, argv, 0);

    spMode   = 1;
    cmdDump(dumpFILE);

    configDump(dumpFILE);
    fprintf(dumpFILE,"\n");

    msgLevel    = 5;
    MsgOverride = 5;
    maxSyslog   = 0;
    MslOverride = 0;
    checkOnly   = 1;

    startLog(-1);
    rtc = readconfigFile (configFile);

    fprintf (stderr, _("Status of config file processing %d\n"),rtc);

    if (msgLevel < 4 && !checkOnly) {
          configDump(dumpFILE);
	  accessDump(dumpFILE);
	  limitsDump(dumpFILE);
    }

    return 0;
}

#endif /* TEST */

#endif /* READCONFIG_C */
